@@ -2,6 +2,7 @@ import argparse
 import os
 import time
 from typing import Dict, Callable
+from shutil import copytree, ignore_patterns, rmtree
 
 import tensorflow as tf
 
@@ -31,9 +32,12 @@ def _train_bert_multitask_keras_model(train_dataset: tf.data.Dataset,
     model_checkpoint_callback = tf.keras.callbacks.ModelCheckpoint(
         filepath=os.path.join(params.ckpt_dir, 'model'),
         save_weights_only=True,
-        monitor='val_loss',
-        mode='min',
-        save_best_only=True)
+        monitor='val_mean_acc',
+        mode='auto',
+        save_best_only=False)
+
+    tensorboard_callback = tf.keras.callbacks.TensorBoard(
+        log_dir=params.ckpt_dir)
 
     with mirrored_strategy.scope():
         model.compile()
@@ -41,9 +45,8 @@ def _train_bert_multitask_keras_model(train_dataset: tf.data.Dataset,
             x=train_dataset,
             validation_data=eval_dataset,
             epochs=params.train_epoch,
-            callbacks=[model_checkpoint_callback],
-            steps_per_epoch=1,
-            validation_steps=1,
+            callbacks=[model_checkpoint_callback, tensorboard_callback],
+            steps_per_epoch=params.train_steps_per_epoch
         )
     model.summary()
 
@@ -56,7 +59,10 @@ def train_bert_multitask(
         params: BaseParams = None,
         problem_type_dict: Dict[str, str] = None,
         processing_fn_dict: Dict[str, Callable] = None,
-        model: tf.keras.Model = None):
+        model: tf.keras.Model = None,
+        create_tf_record_only=False,
+        steps_per_epoch=None,
+        warmup_ratio=0.1):
     """Train Multi-task Bert model
 
     About problem:
@@ -105,24 +111,40 @@ def train_bert_multitask(
         base_dir, dir_name = None, None
     params.train_epoch = num_epochs
     # add new problem to params if problem_type_dict and processing_fn_dict provided
-    if processing_fn_dict:
-        for new_problem, new_problem_processing_fn in processing_fn_dict.items():
-            print('Adding new problem {0}, problem type: {1}'.format(
-                new_problem, problem_type_dict[new_problem]))
-            params.add_problem(
-                problem_name=new_problem, problem_type=problem_type_dict[new_problem], processing_fn=new_problem_processing_fn)
+    if problem_type_dict:
+        params.add_multiple_problems(
+            problem_type_dict=problem_type_dict, processing_fn_dict=processing_fn_dict)
     params.assign_problem(problem, gpu=int(num_gpus),
                           base_dir=base_dir, dir_name=dir_name)
     params.to_json()
 
+    train_dataset = train_eval_input_fn(params)
+    eval_dataset = train_eval_input_fn(params, mode=EVAL)
+    if create_tf_record_only:
+        return
+
+    # get train_steps and update params
+    if steps_per_epoch is not None:
+        train_steps = steps_per_epoch
+    else:
+        train_steps = 0
+        for _ in train_dataset:
+            train_steps += 1
+    params.update_train_steps(train_steps, warmup_ratio=warmup_ratio)
+
+    train_dataset = train_dataset.repeat(params.train_epoch)
+
     mirrored_strategy = tf.distribute.MirroredStrategy()
 
+    if num_gpus > 1:
+        train_dataset = mirrored_strategy.experimental_distribute_dataset(
+            train_dataset)
+        eval_dataset = mirrored_strategy.experimental_distribute_dataset(
+            eval_dataset)
+
     model = _create_keras_model(
         mirrored_strategy=mirrored_strategy, params=params)
 
-    train_dataset = train_eval_input_fn(params)
-    eval_dataset = train_eval_input_fn(params, mode=EVAL)
-
     _train_bert_multitask_keras_model(
         train_dataset=train_dataset,
         eval_dataset=eval_dataset,
@@ -133,6 +155,52 @@ def train_bert_multitask(
     return model
 
 
+def trim_checkpoint_for_prediction(problem: str,
+                                   input_dir: str,
+                                   output_dir: str,
+                                   problem_type_dict: Dict[str, str] = None,
+                                   overwrite=True,
+                                   fake_input_list=None):
+    """Minimize checkpoint size for prediction.
+
+    Since the original checkpoint contains optimizer's variable,
+        for instance, if the use adam, the checkpoint size will 
+        be three times of the size of model weights. This function 
+        will remove those unused variables in prediction to save space.
+
+    Note: if the model is a multimodal model, you have to provide fake_input_list that
+        mimic the structure of real input.
+
+    Args:
+        problem (str): problem
+        input_dir (str): input dir
+        output_dir (str): output dir
+        problem_type_dict (Dict[str, str], optional): problem type dict. Defaults to None.
+        fake_input_list (List): fake input list to create dummy dataset
+    """
+    if overwrite and os.path.exists(output_dir):
+        rmtree(output_dir)
+    copytree(input_dir, output_dir, ignore=ignore_patterns(
+        'checkpoint', '*.index', '*.data-000*'))
+    base_dir, dir_name = os.path.split(output_dir)
+    params = DynamicBatchSizeParams()
+    params.add_multiple_problems(problem_type_dict=problem_type_dict)
+    params.from_json(os.path.join(input_dir, 'params.json'))
+    params.assign_problem(problem, base_dir=base_dir,
+                          dir_name=dir_name, predicting=True)
+
+    model = BertMultiTask(params)
+    if fake_input_list is None:
+        dummy_dataset = predict_input_fn(['fake']*5, params)
+    else:
+        dummy_dataset = predict_input_fn(fake_input_list*5, params)
+    _ = model(next(dummy_dataset.as_numpy_iterator()),
+              mode=tf.estimator.ModeKeys.PREDICT)
+    model.load_weights(os.path.join(input_dir, 'model'))
+    model.save_weights(os.path.join(params.ckpt_dir, 'model'))
+    params.to_json()
+
+
 def eval_bert_multitask(
         problem='weibo_ner',
         num_gpus=1,
@@ -193,7 +261,8 @@ def predict_bert_multitask(
         params: BaseParams = None,
         problem_type_dict: Dict[str, str] = None,
         processing_fn_dict: Dict[str, Callable] = None,
-        model: tf.keras.Model = None):
+        model: tf.keras.Model = None,
+        return_model=False):
     """Evaluate Multi-task Bert model
 
     Available eval_scheme:
@@ -217,15 +286,12 @@ def predict_bert_multitask(
         else:
             base_dir, dir_name = None, None
         # add new problem to params if problem_type_dict and processing_fn_dict provided
-        if processing_fn_dict:
-            for new_problem, new_problem_processing_fn in processing_fn_dict.items():
-                print('Adding new problem {0}, problem type: {1}'.format(
-                    new_problem, problem_type_dict[new_problem]))
-                params.add_problem(
-                    problem_name=new_problem, problem_type=problem_type_dict[new_problem], processing_fn=new_problem_processing_fn)
+        if problem_type_dict:
+            params.add_multiple_problems(
+                problem_type_dict=problem_type_dict, processing_fn_dict=processing_fn_dict)
+        params.from_json(os.path.join(model_dir, 'params.json'))
         params.assign_problem(problem, gpu=1,
-                              base_dir=base_dir, dir_name=dir_name)
-        params.from_json()
+                              base_dir=base_dir, dir_name=dir_name, predicting=True)
     else:
         print('Params problem assigned. Problem list: {0}'.format(
             params.run_problem_list))
@@ -244,6 +310,8 @@ def predict_bert_multitask(
     with mirrored_strategy.scope():
         pred = model.predict(pred_dataset)
 
+    if return_model:
+        return pred, model
     return pred
 
 
